/**************************************************************************
  
  Waldorf Microwave User Programmable Algorithmic Wavetables 
  ( UPAW ) source Assembler
  
  Copyright (c) 1994 Waldorf Electronics GmbH Germany
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details, these can be found 
  in the file "COPYING" that should comes with this program.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Please report bugs and comments to :
  Stefan Stenzel <stenz@waldorf-gmbh.de>
  Waldorf Electronics GmbH Research and Development Department
  Neustrasse 9-12
  D-53498 Waldorf
  Germany

  Waldorf Electronics GmbH Germany hereby disclaims all copyright
  interests in this program.

  UPAW version 1.0 (c) 1994 Stefan Stenzel

**************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#ifndef time_t
#define time_t long
#endif

int parse_line(char *lineptr);	        /* line parser */
int getnum(char *nptr);			/* flimsy expression cracker */
int getlabel(char **lptr);		/* get value of label */
int sendupaw(void);				/* build SysEx string and send over MIDI */
int wlist(char *fname);			/* write listing */
int wfile(char *fname);			/* write SysEx string */


char *Outfilename="UPAW.GSX";	/* Output filename */

/* the extension GSX is taken from the Waldorf Wave's dump
   utility for Generic SysEx Files for compatibility */

int Outflag=0;				/* flag for output file creation */

#ifdef _AMIGA
/*
 *  Amiga Port vor SAS 6.51 by Frank Hoffmann
 *  Port Version 1.0
 */
#include <dos.h> /* for Workbench Arguments */

#include <midi/midibase.h>
#include <midi/midi.h>
#include <clib/midi_protos.h>
#include <pragmas/midi_pragmas.h>

/* AmigaOs Version String */
static char * Version = "$VER: UPaw-Assembler V1.00 Patch V1.00";

/* Amiga Output Function */
int Mstrout(int length, char *mstrptr);

/* Amiga Startup Function */
int AmigaInit(void);

/* Amiga Clean Function */
void AmigaClean(void);

struct MidiBase *MidiBase; /* Shared MIDI Library Base */
struct MSource  *source=0;
struct MRoute   *route=0;

int midistate  = 0;		/* Midi ready ? */
int midioutput = 0;     /* Midi Output wanted ? */
#else

#ifdef ATARI_ST
#include <tos.h>
#define Mstrout Midiws	                /* take Xbios Midi string out function */
#else
#ifdef MSDOS

#include <conio.h>						/* for inp() and outp() */
#include <io.h>							/* for creat() write() and close() */

unsigned int MPUbase=0x330;				/* default MPU base adress */

int Mstrout(int length,char *mstrptr);

int Mstrout(int length,char *mstrptr)
{
	long tcnt;					/* Timeout counter */

	outp(MPUbase+1,0xFF);		/* Send MPU-401 RESET Command */
	inp(MPUbase);	 			/* Dummy read to clear buffer */
	tcnt=0;
	while((inp(MPUbase+1) & 0x040) != 0)
		if(tcnt++ > 60000)
		{
			fprintf(stderr,"MPU initialization failed\n");
			return -1;
		}
	outp(MPUbase+1,0x03F);			/* Set MPU-401 "Dumb UART" Mode */
	inp(MPUbase);   				/* Dummy Read to clear buffer */
	tcnt=0;
	while((inp(MPUbase+1) & 0x080) != 0)  /* Wait for "UART" port ready - */
		if(tcnt++ > 60000)
		{
			fprintf(stderr,"MPU communication failed\n");
			return -1;
		}

	for(; length; length--)
	{
		outp(MPUbase,*mstrptr++);		/* Send Bytes */
		inp(MPUbase);  				    /* Dummy Read to clear buffer */
		tcnt=0;
		while((inp(MPUbase+1) & 0x040) != 0)  /* Wait for "UART" port ready - */
			if(tcnt++ > 60000)
			{
				fprintf(stderr,"MPU communication timeout\n");
				return -1;
			}
	}
	return 0;
}
#else


/*********************************************************************
 Please include your machine specific MIDI string output routine here.
*********************************************************************/
/* dummy prototype */
int Mstrout(int length,char *mstrptr);

/* dummy function */
int Mstrout(int length,char *mstrptr)
{
  Outflag=1;			/* create outputfile instead */
  return 0;
}

#endif
#endif
#endif

int Upaw[64];				/* table to be built */
char Uflag[61];				/* flags for written parameters to detect
								double reference */
char UpawDump[264];     	        /* this one for the SysEx Dump File */
int Wtplace=0;				/* place for wavetable, default 32 */
int DeviceID=127;			/* device ID, default broadcast */
int Devicevalid=0;			/* Commandline option Device marks this */
int errcnt=0;           	        /* error count */

char *Infilename = 0;		/* Input filename */

char *Listfilename;			/* name of listfile */
int Listflag=0;				/* flag for list file creation */

typedef struct {			/* simple structure for fixed labesl */
  char *name;
  int value;
}LABEL;

/*************************************
  Only these fixed labels supported
  See UPAW description for details
**************************************/
LABEL Label[61]={
  "FMFLAG",0,
  "WSFLAG",1,
  "EMFLAG",2,
  "WINDOW",3,
  "FCA",4,
  "FMA",5,
  "MDA",6,
  "FCB",7,
  "FMB",8,
  "MDB",9,
  "T1",10,
  "L1",11,
  "T2",12,
  "L2",13,
  "T3",14,
  "L3",15,
  "T4",16,
  "L4",17,
  "T5",18,
  "L5",19,
  "T6",20,
  "L6",21,
  "T7",22,
  "L7",23,
  "T8",24,
  "L8",25,
  "WA",26,
  "PA",27,
  "DA",28,
  "NA",29,
  "WB",30,
  "PB",31,
  "DB",32,
  "NB",33,
  "WSD",34,
  "P1",35,
  "I1",36,
  "P2",37,
  "I2",38,
  "P3",39,
  "I3",40,
  "P4",41,
  "I4",42,
  "P5",43,
  "I5",44,
  "P6",45,
  "I6",46,
  "P7",47,
  "I7",48,
  "P8",49,
  "I8",50,
  "P9",51,
  "I9",52,
  "P10",53,
  "I10",54,
  "P11",55,
  "I11",56,
  "P12",57,
  "I12",58,
  "P13",59,
  "I13",60
  };


/*************************************************************
main()
parses commandline, reads input file line by line and
calls line parser. If no errors occured, the UPAW is sent
over MIDI or it is written in a file that can be dumped
over MIDI by any dump utility.
**************************************************************/

int main(int argc,char **argv)
{
  FILE *infile;
  int i,c;
  char linebuf[256];		/* buffer to hold input line */

#ifdef _AMIGA
	if(argc == 0)	/* Workbench Arguments */
	{
		argc = _WBArgc;
		argv = _WBArgv;
	}
#endif

  if(argc < 2)
	{
#ifdef _AMIGA
	  fprintf(stderr,"Usage : upaw sourcefile [-dDevicenumber] [-wWavetablenumber]\n             [-lListfile] [-oOutfile] [-m]\n");
#else
#ifdef MSDOS
	  fprintf(stderr,"Usage : UPAW.EXE sourcefile [-dDevicenumber] [-wWavetablenumber]\n                 [-lListfile] [-oOutfile] [-mMPU401-Base-Adress]\n");
#else
#ifdef ATARI_ST
	  fprintf(stderr,"Usage : UPAW.TTP sourcefile [-dDevicenumber] [-wWavetablenumber]\n                 [-lListfile] [-oOutfile]\n");
#else
	  fprintf(stderr,"Usage : upaw sourcefile [-dDevicenumber] [-wWavetablenumber]\n             [-lListfile] [-oOutfile]\n");
#endif
#endif
#endif
	  return(1);
	}

	for (i=1; i<argc; i++)	/* parse commandline */
	{
		if(*argv[i]=='-')
		{
			c=argv[i][1];
			switch(c)
			{
				case 'l':		/* List file */
					Listfilename=&argv[i][2];
					Listflag=1;
					break;
				case 'w':		/* Wavetable number */
					Wtplace=getnum(&argv[i][2]);
					break;
				case 'd':		/* Device ID */
					DeviceID=getnum(&argv[i][2]);
					if(errcnt==0)
					Devicevalid=1;
					break;
				case 'o':		/* Output file */
					if(strlen(&argv[i][2])) Outfilename=&argv[i][2];
					Outflag=1;
					break;
#ifdef _AMIGA
				case 'm':		/* Midi-Output with MIDI Library V2.0 */
					midioutput = 1;
					break;
#endif
#ifdef MSDOS
				case 'm':
					argv[i][1]='$';				/* hack hash to commandline */
					MPUbase=getnum(&argv[i][1]);
					if((MPUbase < 0x200) || (MPUbase > 0x370) || (MPUbase & 0x00F))
					{
						fprintf(stderr,"invalid MPU adress %XH, reset to 330H ( default)\n",MPUbase);
						MPUbase=0x330;
					}
					break;
#endif
					default:
					 fprintf(stderr,"Unknown option : \"%s\"\n",argv[i]);
				break;
			}
		}
		else
		{
			if(!Infilename) /* First String without a leading '-'
							  is Infilename */
			{
				Infilename=argv[i];
				infile=fopen(argv[i],"r");
				if(infile==0)
				{
					fprintf(stderr,"Unble to open file \"%s\"\n",argv[i]);
					return(1);
				}
			}
			else fprintf(stderr,"Unrecognized : \"%s\"\n",argv[i]);
		}
	}
	if(!Infilename)
	{
		fprintf(stderr,"No Infile specified!\n");
		return(1);
	}

#ifdef _AMIGA
	if(midioutput) midistate = AmigaInit();
#endif  

  for(i=0; i<64; i+=2)		/* init UPAW parameters */
    {
      Upaw[i]=0;
      if(i<37)				/* invalidate all parameter indices  */
	Upaw[i+1]=0;
      else
	Upaw[i+1]=-1;
    }
  
  for(i=0; i<61; i++)		/* Init double reference detection flags */
    Uflag[i]=0;
  
  while(NULL != fgets(linebuf,256,infile)) /* read line */
    errcnt+=parse_line(linebuf); 			/* parse line */
  
  if(errcnt == 0)							/* dump table */
    {
      printf("No errors detected\n");
      sendupaw();
    }  
  else
    {
      printf("%d error(s) detected\n");
      return errcnt;							/* stop here in case of errors */	
    }
  if(Listflag)
    wlist(Listfilename);
  
  if(Outflag)
    wfile(Outfilename);

#ifdef _AMIGA  
	if(midioutput) AmigaClean(); 
#endif

  return(errcnt);
}

/*************************************************************
  parse_line()
  expects following format :
  [index:value][;comment]
  both fields are of course optional.
  There are two special keywords that are treated special :
  "Device" : must be followed by a number and sets the global
  variable DeviceID. 
  "Wavetable" : must be followed by a number and sets the global
  variable Wtplace.
  
  However, if one of these has already been given on the commandline,
  it will be ignored, thus commandline options take priority.
  
  returns 1 if an error occured, else 0.
***************************************************************/

int parse_line(char *lineptr)
{
  int pindex,pvalue;
  char *pvalptr,*cptr,c;
  
  cptr=lineptr;
  
  while((c=*cptr)!=0)		/* toupper() performs tolower for upper case at least
				 on my Un*x system, so here comes a workaround */
    {
      if((c >= 'a') && (c <= 'z'))
	*cptr=c-32;
      cptr++;
    }
  
  if((cptr=strchr(lineptr,';'))!=NULL)
    *cptr=0;			/* terminate string here */
  
  while((*lineptr == ' ') || (*lineptr == 9))
    lineptr++;
  
  if((*lineptr == 0) || 
     (*lineptr==10) || 
     (*lineptr==13))	/* nothing in this line */
    return 0;
  
  if(strncmp("WAVETABLE",lineptr,9)==0)
    {
      if(Wtplace==0)	/* commandline has priority ! */
	Wtplace=getnum(lineptr+9);
      return 0;
    }		
  if(strncmp("DEVICE",lineptr,6)==0)
    {
      if(Devicevalid==0)	/* commandline has priority ! */
	DeviceID=getnum(lineptr+6);
      return 0;
    }
  if((pvalptr=strchr(lineptr,':'))==NULL)	/* crack operands */
    {
      fprintf(stderr,"':' expected \n");
      return(1);
    }
  *pvalptr++=0;		/* replace ':' by zero */
  
  pindex = getnum(lineptr);
  pvalue = getnum(pvalptr);  
  if(pindex > 60)
    {
      fprintf(stderr,"index value %d out of range\n",pindex);
      return 1;
    }
  if(Uflag[pindex])			/* if this has already been written */
    {
      fprintf(stderr,"Attempt to write value for \"%s\" twice\n",Label[pindex].name);
      return 1;
    }	
  Uflag[pindex]=1;			/* mark as written */
  Upaw[pindex+2]=pvalue;	/* write value */
  return 0;
}

/*********************************************************************
  getnum()
  This is the flimsiest mathematical expression parser I have ever seen,
  but nevertheless it supports the fixed labels, hexadecimal and decimal
  numbers, fixpoint numbers, and the basic mathematical operators.
  The priorities of operators are completely ignored, and, even worse,
  the operations preceed from right to left.
  If an error is detected, the global variable errcnt is increased and
  the result will be zero or garbage.
**********************************************************************/

int getnum(char *nptr)
{
  int num=0,frac=0,fpflag=0,ffact=10,c,step=0,hexflag=0;
  
  while((c=*nptr) != 0)
    {
      switch(c)
	{
	case ' ':
	case 9:
	  step--;		/* hack ! ( : do not step here ) */
	  break;		/* skip whitespace */
	  
	case 13:		/* End of line */
	case 10:
	  return(num);
	  
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	  if(hexflag)
	    {
	      num<<=4;
	      num+=c-'0';
	    }	
	  else
	    {
	      if(fpflag == 0)	/* if this is not a fixpoint number */
	    	{
		  num*=10;
		  num+=c-'0';		/* poor mans atoi */
	    	}	  	else if(fpflag==1)
		  {
		    frac+=((c-'0')<<11)/ffact; /* update fractional part */
		    ffact*=10;	/* update fractional divisor */
		    num&=0xFE00;	/* delete old fractional part */
		    num|=(frac+2)>>2;	/* set new fractional part */
		  }
	    }	
	  break;
	case '.':		/* oh my god, a fixpoint number ! */
	  if(hexflag)
	    {
	      fprintf(stderr,"invalid number \"%s\"\n",nptr);
	      errcnt++;
	      return(0);
	    }
	  num<<=9;		/* integer part to upper 7 bits */
	  fpflag++;
	  break;
	  
	case '+':		/* recurse for operations */
	  return(num+getnum(++nptr));
	case '-':
	  return(num-getnum(++nptr));
	case '*':
	  return(num*getnum(++nptr));
	case '/':
	  if((c=getnum(++nptr))!=0)
	    return(num/c);
	  else
	    {
	      fprintf(stderr,"Attempt to divide by zero : \"%s\"\n",nptr);
	      errcnt++;
	      return(0);
	    }
	case '&':
	  return(num&getnum(++nptr));
	case '|':
	  return(num|getnum(++nptr));
	case '^':
	  return(num^getnum(++nptr));
	case '$':
	  if(step==0)
	    hexflag=1;
	  else
	    {
	      fprintf(stderr,"invalid number \"%s\"\n",nptr);
	      errcnt++;
	      return(0);
	    }				  
	  break;
	case 'A':
	case 'B':
	case 'C':
	case 'D':
	case 'E':
	case 'F':
	  if(hexflag)
	    {
	      num<<=4;
	      num+=c-'A';
	      break;
	    }				
	default:
	  if(step==0)
	    num=getlabel(&nptr);
	  else
	    {
	      fprintf(stderr,"Invalid number : \"%s\"\n",nptr);
	      errcnt++;
	      return(0);
	    }
	}
      nptr++;
      step++;
    }
  return(num);
}

/******************************************************************
  getlabel()
  returns the numerical value of a label.
  if the label cannot be found, it returns zero and the global
  variable errcnt is incremented.
*******************************************************************/

int getlabel(char **lptr)
{
  int i,length;
  
  for(i=0; i<61; i++)
    {
      if(strncmp(Label[i].name,*lptr,length=(int)strlen(Label[i].name))==0)
	{
	  *lptr+=length-1;	/* increase pointer */
	  return(Label[i].value);
	}
    }
  fprintf(stderr,"Unknown expression : \"%s\"\n",*lptr);
  errcnt++;
  return 0;
}

/********************************************************************
  Creates System Exclusive MIDI string according to the global variables
  DeviceID, Wtplace, and Upaw.
  returns -1 if an errror is detected, otherwise zero.
  On some systems, like Atari ST computers, the SysEx string will be
  sent immediately over MIDI.
*********************************************************************/

int sendupaw(void)
{
  int i,val;
  char *dumptr,cs,c;
  
  Upaw[0]=0x12DE;	/* set magic long */
  Upaw[1]=0xC0DE;
  
  dumptr=UpawDump;
  *dumptr++=0xF0;	/* SysEx */
  *dumptr++=0x3E;	/* Waldorf Electronics Manufacturer ID */
  *dumptr++=0;	/* Microwave */
  DeviceID&=0x007F;	/* clear bit 7 for security */
  *dumptr++=DeviceID;	/* Device ID */
  *dumptr++=0x45;	/* Control Table Dump */
  if((Wtplace < 33) || (Wtplace > 56))
    {
      fprintf(stderr,"invalid Wavetable number %d\n",Wtplace);
      return -1;
    }
  cs=Wtplace-1;	/* checksum starts here */
  *dumptr++=Wtplace-1;
  for(i=0; i<64; i++)
    {
      val=Upaw[i];
      c=val>>12;	/* upper 4 bits */
      c&=0xF;		/* for security */
      cs+=c;
      *dumptr++=c;
      c=val>>8;	/* next 4 bits */
      c&=0xF;		/* mask away superfluid bits */
      cs+=c;
      *dumptr++=c;
      c=val>>4;	/* next 4 bits */
      c&=0xF;		/* mask away superfluid bits */
      cs+=c;
      *dumptr++=c;
      c=val;	    /* lower 4 bits */
      c&=0xF;		/* mask away superfluid bits */
      cs+=c;
      *dumptr++=c;
    }	
  cs&=0x7F;		/* checksum to 7 bits */
  *dumptr++=cs;
  *dumptr=0xF7;	/* End of SysEx or EOX */
  
  Mstrout(264,UpawDump);
  return 0;
}	

/***************************************************************** 
  wlist()
  write complete parameter listing to file 
  The list format is as follows :
  -Header, including copyright,date, and filename,
  -Device and Wavetable information
  -for each parameter :
  Label   : $Hexvalue ; decimal value [/ fixpoint value][/ Label]
  e.g.:
  
  FCA     : $0200  ; 512 / 1.000
  ...
  P1      : $0004  ;   4 / FCA
  
  The listing can be fed back to the assembler as a source.
  Expects complete parameter list in Upaw[].
******************************************************************/

int wlist(char *fname)
{
  FILE *outfile;
  int i,val;
  double fixp;
  time_t now;

  outfile=fopen(fname,"w");
  if(outfile==NULL)
	{
	  fprintf(stderr,"could not create \"%s\"\n",fname);
	  return -1;
	}
  fprintf(outfile,"; UPAW Assembler V 1.0\n; (c) 1994 Waldorf Electronics GmbH Germay\n");
  fprintf(outfile,"; GNU General Public License rules apply\n");
  time(&now);
  fprintf(outfile,"; Date : %s",ctime(&now));
  fprintf(outfile,"; File : \"%s\"\n\n",Infilename);
  fprintf(outfile,"Device    $%X\n",DeviceID);
  fprintf(outfile,"Wavetable %d\n\n",Wtplace);
  for(i=0; i<61; i++)
	{
	  val=Upaw[i+2];

	  fixp=((double)val)/512.0;
	  if((i < 35) || ((i&0x0001) == 0))
	/* if not a parameter index */
	fprintf(outfile,"%-7s: $%04X  ; %5d / % 2.3f\n",Label[i].name,val&0xFFFF,val,fixp);
	  else
	{
	  if((val <= 60) && (val >= 0))	/* if a valid parameter is selected */
		fprintf(outfile,"%-7s: $%04X  ; %5d / %s\n",Label[i].name,val&0xFFFF,val,Label[val].name);
	  else
		fprintf(outfile,"%-7s: $%04X  ; %5d\n",Label[i].name,val&0xFFFF,val);
	}				
    }
  fclose(outfile);
  return 0;
}

/************************************************************
  wfile()
  writes SysEx string to file			
  Expects complete SysEx string in UpawDump[].
************************************************************/

int wfile(char *fname)	
{
  int sxfile;
  
  sxfile=creat(fname,660);					/* create file */
  if(sxfile <= 0)
    {
      fprintf(stderr,"could not create \"%s\"\n",fname);
      return sxfile;
    }		
  
  if(write(sxfile,UpawDump,264L)!=264)	/* write data */
    {
      fprintf(stderr,"Write error on output file \"%s\"\n",fname);
      close(sxfile);
      return -1;
    }		
  
  close(sxfile);							/* close file */
  
  return 0;
}	

#ifdef _AMIGA

int AmigaInit(void)
/* Open MIDI-Resources */ 
{
	if (!(MidiBase = (struct MidiBase*)OpenLibrary (MIDINAME,MIDIVERSION))) 
	{
		printf ("can't open midi.library\n");
		return -1;
	}
	/* Create Source */
	if (!(source = CreateMSource (NULL,NULL))) 
	{
		printf ("can't create MIDI-Source\n");
		return -1;	
	}
	/* Conect Source to Standard-MidiOut */ 
	if (!(route = MRouteSource (source, "MidiOut", NULL)))
	{
		printf ("can't create MIDI-Route (can't find midioutput?)\n");
		return -1;
	}
}

void AmigaClean(void)
/* Clean MIDI Resources */
{
	if (route)    DeleteMRoute(route);
    if (source)   DeleteMSource(source);
    if (MidiBase) CloseLibrary((struct Library*)MidiBase);
}

int Mstrout(int length,char *mstrptr)
/* Send MIDI Stream if desired */
{	
	if(midistate && midioutput) PutMidiStream(source,NULL,mstrptr,length,length);
	return 0;
}
#endif